import gzip
import io
import json
import os
import pickle
import subprocess
from dataclasses import dataclass
from datetime import datetime
from glob import glob
from pathlib import Path
from typing import Iterable, List, Optional, Self
from urllib.parse import quote, unquote, urlparse

import numpy as np
import pandas as pd
from boto3.s3.transfer import TransferConfig
from tqdm import tqdm

from .client import get_client as _get_client
from .config import YeS3Config
from .utils.decorators import timeit_opt
from .utils.logs import get_logger

log = get_logger()
_client = _get_client()

# Don't overwrite the value of YES3_CONFIG if it has already been set
global YES3_CONFIG  # noqa: F824
try:
    YES3_CONFIG
except NameError:
    YES3_CONFIG = YeS3Config()

# TODO:
#  1. Replace _verbose_print with logger
#  2. Add overwrite protection options
#  3. Add some more shortcut methods to S3Location, perhaps some aliases


def config(**config_params):
    global YES3_CONFIG  # noqa: F824
    for param, value in config_params.items():
        setattr(YES3_CONFIG, param, value)
    return YES3_CONFIG


def is_s3_url(s: str) -> bool:
    return s.startswith('s3://') or s.startswith('https://s3.')


class S3Location:
    def __init__(self, bucket: str | Self, key: Optional[str] = None, region: Optional[str] = None):
        if isinstance(bucket, S3Location):
            loc = bucket
            if key:
                loc = loc.join(key)
            self.bucket = loc.bucket
            self.key = loc.key
        elif is_s3_url(bucket):
            loc = type(self).parse(bucket)
            if key is not None:
                loc = loc.join(key)
            self.bucket = loc.bucket
            self.key = loc.key
        else:
            self.bucket = bucket
            self.key = key
        self.region = region

    def __repr__(self) -> str:
        params = {"bucket": self.bucket, "key": self.key}
        if self.region:
            params["region"] = self.region
        return f"{type(self).__name__}({', '.join([f'{k}={v}' for k, v in params.items()])})"

    def __eq__(self, other) -> bool:
        if isinstance(other, S3Location):  # not using type(self) in case comparing different subclasses of S3Location
            return ((self.bucket == other.bucket) and
                    ((self.key is None and other.key is None) or (self.key.rstrip('/') == other.key.rstrip('/'))))
        elif isinstance(other, str):
            return self == type(self).parse(other)
        else:
            return False

    @property
    def s3_uri(self) -> str:
        s = f's3://{self.bucket}'
        if self.key is not None:
            s += f'/{self.key}'
        return s

    @property
    def https_url(self, region=None) -> str:
        region = region or (self.region or YES3_CONFIG.default_region)
        s = f'https://s3.{region}.amazonaws.com/{quote(self.bucket)}'
        if self.key is not None:
            s += f'/{quote(self.key)}'
        return s

    @classmethod
    def parse(cls, url: str) -> Self:
        parsed_url = urlparse(url)
        if parsed_url.scheme == 's3':
            bucket = unquote(parsed_url.netloc)
            key = unquote(parsed_url.path[1:])
            region = None
        elif parsed_url.scheme.startswith('http'):
            bucket, *key_parts = unquote(parsed_url.path[1:]).split('/')
            key = '/'.join(key_parts)
            region = parsed_url.netloc.split('.')[1]
        else:
            raise ValueError(f'Unrecognized URL scheme: {parsed_url.scheme}')
        return cls(bucket=bucket, key=key, region=region)

    def join(self, *parts) -> Self:
        key_parts = self.key.split('/') if self.key else []
        if any(not isinstance(part, str) or is_s3_url(part) for part in parts):
            raise TypeError('All arguments to `join` must be non-url strings.')
        for part in parts:
            key_parts += part.split('/')
        new_key = '/'.join(key_parts)
        while '//' in new_key:
            new_key = new_key.replace('//', '/')
        return type(self)(self.bucket, new_key, self.region)

    def exists(self) -> bool:
        return len(list_objects(self, limit=1)) > 0

    def is_bucket(self) -> bool:
        return self.key is None or len(self.key) == 0

    def is_object(self) -> bool:
        objects = list_objects(self, limit=2)
        return len(objects) == 1 and objects[0].key == self.key

    def is_dir(self) -> bool:
        if not self.exists():
            return False
        if self.is_object():
            return False
        first_obj = list_objects(self, limit=1)[0]
        suffix = first_obj.s3_uri.split(self.s3_uri.rstrip('/'))[1]
        if suffix.startswith('/'):
            return True
        return False

    def is_dir_path(self) -> bool:
        if self.is_bucket():
            return True
        if self.key.endswith('/'):
            return True
        return self.is_dir()

    def split_key(self) -> tuple[str, str]:
        if self.key:
            key = self.key.rstrip('/')
            if '/' not in key:
                return '', key
            else:
                return tuple(key.rsplit('/', maxsplit=1))
        else:
            return '', ''

    @property
    def parent(self) -> Self:
        parent_key = self.split_key()[0]
        if not parent_key:
            parent_key = None
        return type(self)(self.bucket, parent_key, self.region)

    def get_object_metadata(self) -> 'S3Object' | List['S3Object']:
        objects = list_objects(self, return_metadata=True)
        if len(objects) == 1 and objects[0].location.key == self.key:
            return objects[0]
        else:
            return objects

    def __truediv__(self, other):
        return self.join(other)


S3LocationLike = str | S3Location
LocalPathLike = str | os.PathLike


def get_size_str(bytes: int) -> str:
    if bytes < 1024:
        return f'{bytes} B'
    elif bytes < 1024 ** 2:
        return f'{bytes / 1024:.2f} KB'
    elif bytes < 1024 ** 3:
        return f'{bytes / 1024 ** 2:.2f} MB'
    else:
        return f'{bytes / 1024 ** 3:.2f} GB'


@dataclass
class S3Object:
    location: S3Location
    last_modified: datetime
    e_tag: str
    size: int
    storage_class: str

    @classmethod
    def from_dict(cls, bucket: str, d: dict) -> Self:
        return cls(
            location=as_s3_location(bucket, d['Key']),
            last_modified=d['LastModified'],
            e_tag=d['ETag'],
            size=d['Size'],
            storage_class=d['StorageClass'],
        )

    def __repr__(self) -> str:
        params = [
            self.location.s3_uri,
            get_size_str(self.size),
            self.last_modified.isoformat(),
        ]
        return f"{type(self).__name__}({', '.join(params)})"


@dataclass
class S3Prefix:
    location: S3Location
    key_count: int = -1  # -1 means unknown number of objects with this prefix, 0 means no objects have this prefix

    def count_objects(self) -> int:
        objs = list_objects(self.location)
        self.key_count = len(objs)
        return self.key_count

    def __repr__(self) -> str:
        params = [self.location.s3_uri]
        if self.key_count >= 0:
            params.append(f'{self.key_count} objects')
        return f"{type(self).__name__}({', '.join(params)})"


def as_s3_location(bucket_or_location: S3LocationLike, key: Optional[str] = None) -> S3Location:
    if isinstance(bucket_or_location, S3Location):
        location = bucket_or_location
        if key is not None:
            location.key = key
    elif bucket_or_location.startswith('s3://') or bucket_or_location.startswith('https://'):
        location = S3Location.parse(bucket_or_location)
        if key is not None:
            location = location.join(key)
    else:
        location = S3Location(bucket_or_location, key)
    return location


def list_objects(
        bucket_or_location: S3LocationLike,
        prefix: Optional[str] = None,
        limit: Optional[int] = None,
        return_metadata: bool = False,
) -> list[S3Location | S3Object]:
    location = as_s3_location(bucket_or_location, prefix)

    def get_next_page(cont_token=None) -> [Optional[str], list[S3Object]]:
        args = dict(Bucket=location.bucket, Prefix=location.key)
        if limit is not None and limit >= 0:
            args['MaxKeys'] = int(limit)
        if cont_token is not None:
            args['ContinuationToken'] = cont_token
        resp = _client.list_objects_v2(**args)
        next_token = resp.get('NextContinuationToken')
        contents = resp.get('Contents', [])
        parsed_contents = [S3Object.from_dict(location.bucket, d) for d in contents]
        return next_token, parsed_contents

    token, results = get_next_page()
    while token is not None and (limit is None or len(results) < limit):
        token, next_results = get_next_page(token)
        results += next_results

    if limit is not None and limit >= 0:
        results = results[:limit]

    if return_metadata:
        return results
    else:
        return [r.location for r in results]


@timeit_opt
def list_dir(
        bucket_or_location: S3LocationLike,
        prefix: Optional[str] = None,
        depth=None,
        recursive=False,
        limit: Optional[int] = None,
        return_metadata: bool = False,
        count_only: bool = False,
) -> list[S3Location | S3Object | S3Prefix] | int:
    location = as_s3_location(bucket_or_location, prefix)
    if location.is_dir_path():
        location = location.join('')

    if depth is None:
        if recursive:
            depth = -1
        else:
            depth = 1

    paginator = _client.get_paginator('list_objects_v2')
    paginator_args = {'Bucket': location.bucket, 'Prefix': location.key, 'Delimiter': '/'}
    if limit is not None and limit >= 0:
        paginator_args['PaginationConfig'] = {'MaxItems': int(limit)}

    if count_only:
        results = 0
    else:
        results = []
    if depth == 0:
        return results

    for page in paginator.paginate(**paginator_args):
        # CommonPrefixes and Contents might not be included in a page if there
        # are no items, so use .get() to return an empty list in that case
        for item in page.get('CommonPrefixes', []):
            if count_only:
                results += 1
            else:
                results.append(S3Prefix(S3Location(location.bucket, item['Prefix'])))
                if depth > 1 or depth < 0:
                    results.extend(list_dir(location.bucket, item['Prefix'], depth - 1, return_metadata=True))
        for item in page.get('Contents', []):
            if count_only:
                results += 1
            else:
                results.append(S3Object.from_dict(location.bucket, item))

    if limit is not None and limit >= 0:
        if count_only:
            return min(results, limit)
        else:
            results = results[:limit]

    if count_only or return_metadata:
        return results
    else:
        return [r.location for r in results]


def PathExt(*args, **kwargs) -> os.PathLike:
    p = Path(*args, **kwargs)
    cls = type(p)

    class UnmadeDirPath(cls):
        _is_unmade_dir = True

    if not p.exists() and str(args[-1]).endswith('/'):
        return UnmadeDirPath(p)
    else:
        return p


def is_unmade_dir(path: os.PathLike) -> bool:
    p = PathExt(path)
    try:
        return p._is_unmade_dir
    except AttributeError:
        return False


def _parse_progress_arg(progress_arg) -> tuple[str, int | float]:
    if isinstance(progress_arg, str):
        progress_mode = YES3_CONFIG.check_progress_mode(progress_arg)
        progress_size_threshold = YES3_CONFIG.progress_size_threshold
    elif progress_arg is False:
        progress_mode = YES3_CONFIG.check_progress_mode('off')
        progress_size_threshold = YES3_CONFIG.progress_size_threshold
    elif progress_arg is None:
        progress_mode = YES3_CONFIG.progress_mode
        progress_size_threshold = YES3_CONFIG.progress_size_threshold
    else:
        try:
            progress_size_threshold = float(progress_arg)
            progress_mode = YES3_CONFIG.check_progress_mode('large')
        except (ValueError, TypeError):
            raise ValueError(f'Invalid progress arg value: {progress_arg}')

    return progress_mode, progress_size_threshold


def _get_upload_prog_callback(progress_arg, filesize: int, path: Path):
    # Parse progress arg
    if callable(progress_arg):
        return progress_arg
    progress_mode, progress_size_threshold = _parse_progress_arg(progress_arg)

    if progress_mode == 'off':
        return

    if progress_mode == 'large' and filesize < progress_size_threshold:
        return

    pbar = tqdm(total=filesize, desc=f"Uploading {path.name}")

    def f(size):
        pbar.update(size)

    return f


def _upload_file(local_path: LocalPathLike, location: S3Location, progress=None) -> S3Location:
    local_path = Path(local_path).resolve()
    if not local_path.exists():
        raise FileNotFoundError(str(local_path))
    if location.is_dir_path():
        filename = local_path.name
        location = location.join(filename)
    log.info(f'Uploading {local_path} to {location.s3_uri}... ')
    filesize = local_path.stat().st_size
    transfer_config = YES3_CONFIG.upload_transfer_config or TransferConfig()
    if filesize < transfer_config.multipart_threshold:
        with open(local_path, 'rb') as f:
            _client.put_object(Bucket=location.bucket, Key=location.key, Body=f)
    else:
        opt_prog_callback = _get_upload_prog_callback(progress, filesize, local_path)
        _client.upload_file(
            Filename=str(local_path),
            Bucket=location.bucket,
            Key=location.key,
            Config=transfer_config,
            Callback=opt_prog_callback,
        )
    log.info(f'Uploaded {local_path} to {location.s3_uri}')
    return location


def _highest_common_dir(paths: Iterable[LocalPathLike], resolve=True) -> os.PathLike:
    if resolve:
        paths = [Path(p).resolve() for p in paths]
    else:
        paths = [Path(p) for p in paths]
    if len(paths) == 0:
        raise ValueError('At least one path is required')
    next_path = paths.pop(0)
    common_dir = next_path.parent
    while paths:
        next_path = paths.pop(0)
        while not str(next_path).startswith(str(common_dir)):
            common_dir = common_dir.parent
    return common_dir


def upload(
        local_path: LocalPathLike,
        bucket_or_location: S3LocationLike,
        prefix: Optional[str] = None,
        recursive: bool = False,
        base_dir=None,
        progress=None,
        # workers: int = 1,
        # threads: int = 1,
) -> S3Location | list[S3Location]:
    local_path = Path(local_path).resolve()
    location = as_s3_location(bucket_or_location, prefix)

    if not local_path.exists():
        raise FileNotFoundError(f'No paths found matching `{local_path}`')
    elif local_path.is_file():
        return _upload_file(local_path, location, progress=progress)
    else:
        if not recursive:
            raise ValueError(f'Must set `recursive=True` for directories or paths with wildcards (*): {local_path}')

        if '*' in str(local_path):
            local_paths = [Path(p) for p in glob(str(local_path))]
        elif local_path.is_dir():
            local_paths = []
            for (dir_path, _, filenames) in os.walk(local_path):
                for filename in filenames:
                    local_paths.append(Path(dir_path) / filename)
        else:
            local_paths = []

        if len(local_paths) == 0:
            raise FileNotFoundError(f'No paths found matching `{local_path}`')

        common_dir = _highest_common_dir(local_paths)
        if base_dir is None:
            base_dir = common_dir
        else:
            base_dir = Path(base_dir).resolve()
            if not str(common_dir).startswith(str(base_dir)):
                raise ValueError(f'Local files are not in subdirectories of the given base dir `{base_dir}`')

        # if workers == -1 or workers > 1:
        #     rel_paths = [location.join(str(p.relative_to(base_dir))) for p in local_paths]
        #     with Pool(workers) as pool:
        #         locations = pool.map(_upload_file, local_paths, rel_paths)
        # elif threads > 1:
        #     rel_paths = [location.join(str(p.relative_to(base_dir))) for p in local_paths]
        #     with ThreadPoolExecutor(threads) as pool:
        #         locations = list(pool.map(_upload_file, local_paths, rel_paths))
        # else:
        locations = []
        for path in local_paths:
            rel_path = path.relative_to(base_dir)
            loc = _upload_file(path, location.join(str(rel_path)), progress=progress)
            locations.append(loc)

        return locations


def _get_download_prog_callback(progress_arg, location: S3Location):
    if callable(progress_arg):
        return progress_arg
    progress_mode, progress_size = _parse_progress_arg(progress_arg)

    if progress_mode == 'off':
        return

    obj_meta = list_objects(location, return_metadata=True)[0]
    obj_size = obj_meta.size
    if progress_mode == 'large' and obj_size < progress_size:
        return

    pbar = tqdm(total=obj_size, desc=f"Downloading {obj_meta.location.key.rsplit('/', 1)[-1]}")

    def f(size):
        pbar.update(size)

    return f


def _download_object(location: S3Location, local_path: LocalPathLike, progress=None) -> os.PathLike:
    input_local_path = local_path
    local_path = Path(local_path).resolve()
    if local_path.is_dir() or is_unmade_dir(input_local_path):
        os.makedirs(local_path, exist_ok=True)
        filename = location.split_key()[1]
        local_path = local_path / filename
    else:
        os.makedirs(local_path.parent, exist_ok=True)
    log.info(f'Downloading {location.s3_uri} to {local_path}... ')
    callback = _get_download_prog_callback(progress, location)
    _client.download_file(location.bucket, location.key, str(local_path), Callback=callback)
    log.info(f'Downloaded {location.s3_uri} to {local_path}')
    return local_path


def download(
        bucket_or_location: S3LocationLike,
        prefix: str | LocalPathLike,
        local_path: Optional[LocalPathLike] = None,
        recursive: bool = False,
        base_dir=None,
        progress=None,
) -> str | list[str]:
    if local_path is None:
        local_path = prefix
        prefix = None

    location = as_s3_location(bucket_or_location, prefix)

    if is_unmade_dir(local_path):
        os.makedirs(local_path)
    local_path = Path(local_path).resolve()

    if not location.exists():
        raise FileNotFoundError(f'No object(s) present at {location.s3_uri}')
    elif not location.is_object() and not recursive:
        raise ValueError(f'{location.s3_uri} is not a uri for a single object; '
                         'use `recursive=True` to download all objects with this prefix')

    if location.is_object():
        p = _download_object(location, local_path, progress=progress)
        return str(p)
    else:
        objects = list_objects(location)
        common_dir = _highest_common_dir([loc.key for loc in objects], resolve=False)
        if base_dir is None:
            base_dir = common_dir
        else:
            base_dir = Path(base_dir).resolve()
            if not str(common_dir).startswith(str(base_dir)):
                raise ValueError(f'Local files are not in subdirectories of the given base dir `{base_dir}`')

        local_paths = []
        for loc in objects:
            rel_dir = Path(loc.key).relative_to(base_dir).parent
            rel_path = local_path / str(rel_dir) / loc.split_key()[1]
            p = _download_object(loc, rel_path, progress=progress)
            local_paths.append(str(p))
        return local_paths


def _delete_object(location: S3Location):
    log.info(f'Deleting {location.s3_uri}... ')
    _client.delete_object(Bucket=location.bucket, Key=location.key)
    log.info(f'Deleted {location.s3_uri}')


def delete(
        bucket_or_location: S3LocationLike,
        prefix: Optional[str] = None,
        recursive: bool = False,
):
    location = as_s3_location(bucket_or_location, prefix)
    if not location.exists():
        raise FileNotFoundError(f'No object(s) at {location.s3_uri}')
    elif not location.is_object():
        if not recursive:
            raise ValueError(f'Multiple objects with prefix {location.s3_uri}, set `recursive=True` to delete them all')
        for loc in list_objects(location):
            _delete_object(loc)
    else:
        _delete_object(location)


def read(
        bucket_or_location: S3LocationLike,
        prefix: Optional[str] = None,
        file_type: Optional[str] = None,
        local_temp_file=None,
        progress=None,
        **kwargs,
):
    def read_body(body):
        if file_type == 'pkl':
            return pickle.load(body, **kwargs)
        elif file_type == 'json':
            return json.load(body, **kwargs)
        elif file_type == 'csv':
            return pd.read_csv(body, **kwargs)
        elif file_type == 'npy':
            return np.load(body, **kwargs)
        elif file_type == 'parquet':
            return pd.read_parquet(body, **kwargs)
        elif file_type in ('txt', 'text'):
            return body.read().decode()
        else:
            return body

    location = as_s3_location(bucket_or_location, prefix)
    if not location.exists():
        raise FileNotFoundError(f'No object(s) present at {location.s3_uri}')
    elif not location.is_object():
        raise ValueError(f'{location.s3_uri} is not a uri for a single object, cannot use `read` method.')

    if callable(progress):
        with_progress = True
    else:
        if progress is False:
            with_progress = False
        else:
            progress_mode, progress_size = _parse_progress_arg(progress)
            if progress_mode == 'off':
                with_progress = False
            elif progress_mode == 'all':
                with_progress = True
            else:
                obj_size = list_objects(location, limit=1, return_metadata=True)[0].size
                with_progress = (obj_size >= progress_size)

    ext = Path(location.key).suffix
    if file_type is None and ext:
        file_type = ext.lstrip('.')
    if file_type is not None:
        file_type = file_type.lstrip('.').lower()

    if with_progress or local_temp_file:
        if not local_temp_file:
            local_temp_file = f"TMPFILE.{datetime.now().strftime('%Y%m%dT%H%M%S.%f')}"
        log.info(f"Reading to local temp {file_type} file: {local_temp_file}")
        body = download(location, local_temp_file, progress=progress)
        with open(body, 'rb') as f:
            obj = read_body(f)
        log.info(f'Deleting {local_temp_file}')
        os.remove(local_temp_file)
        return obj
    else:
        body = _client.get_object(Bucket=location.bucket, Key=location.key)['Body']
        return read_body(body)


def write_to_s3(
        obj,
        bucket_or_path: str | S3Location,
        key: Optional[str] = None,
        local_temp_file=None,
        file_type: Optional[str] = None,
        progress=None,
        **kwargs
) -> S3Location:
    s3_loc = S3Location(bucket_or_path, key)

    if not local_temp_file:
        local_temp_file = f"TMPFILE.{datetime.now().strftime('%Y%m%dT%H%M%S.%f')}"

    if file_type is None:
        file_type = os.path.splitext(s3_loc.key)[1].lstrip('.').lower()
    if not file_type:
        file_type = 'pkl'

    log.info(f"Writing local temp {file_type} file: {local_temp_file}")
    if file_type == 'json':
        with open(local_temp_file, 'w') as f:
            json.dump(obj, f, **kwargs)
    elif file_type in ('pkl', 'pickle'):
        with open(local_temp_file, 'wb') as f:
            pickle.dump(obj, f, **kwargs)
    elif file_type == 'parquet':
        obj.to_parquet(local_temp_file, **kwargs)
    elif file_type in ('gz', 'gzip'):
        with open(local_temp_file, 'wb') as f:
            pickle.dump(io.BytesIO(gzip.compress(obj)), f, **kwargs)
    elif file_type == 'txt':
        with open(local_temp_file, 'w') as f:
            f.write(str(obj))
    else:
        raise ValueError(f"Unrecognized file_type: '{file_type}'")

    written_loc = upload(local_temp_file, s3_loc, progress=progress)
    log.info(f'Deleting {local_temp_file}')
    os.remove(local_temp_file)
    return written_loc


def touch(bucket_or_location: S3LocationLike, prefix: Optional[str] = None):
    location = as_s3_location(bucket_or_location, prefix)
    _client.put_object(Bucket=location.bucket, Key=location.key, Body=b'')
    log.info(f'Created empty object at {location.s3_uri}')


@timeit_opt(default=True)
def large_recursive_delete(bucket_or_location: S3LocationLike, prefix: Optional[str] = None, timeit=True):
    """See https://serverfault.com/a/1123717"""
    location = as_s3_location(bucket_or_location, prefix)
    if not location.exists():
        raise FileNotFoundError(f'No object(s) at {location.s3_uri}')
    elif location.is_object():
        raise ValueError(f'{location.s3_uri} is an object, not a prefix; use `delete` instead')
    else:
        cmd = ("""aws s3api list-objects-v2 --bucket {bucket} --prefix {prefix} --output text --query """
               """'Contents[].[Key]' | grep -v -e "'" | tr '\\n' '\\0' | xargs -0 -P2 -n500 bash -c """
               """'aws s3api delete-objects --bucket {bucket} --delete """
               """"Objects=[$(printf "{{Key=%q}}," "$@")],Quiet=true"' _ """
               ).format(bucket=location.bucket, prefix=location.key)
    if timeit:
        print(f'Starting recursive delete with prefix {location.s3_uri} at {datetime.now().isoformat()}')
    os.system(cmd)


@timeit_opt(default=True)
def list_many_objects(bucket_or_location: S3LocationLike, prefix: Optional[str] = None, timeit=True) -> list[str]:
    location = as_s3_location(bucket_or_location, prefix)
    if not location.exists():
        raise FileNotFoundError(f'No object(s) at {location.s3_uri}')
    elif location.is_object():
        return [location.s3_uri]
    else:
        cmd = ("aws s3api list-objects-v2 --bucket {bucket} --prefix {prefix} --output text --query 'Contents[].[Key]'"
               ).format(bucket=location.bucket, prefix=location.key)
    if timeit:
        print(f'Starting list-objects call with prefix {location.s3_uri} at {datetime.now().isoformat()}')
    output = subprocess.run(cmd, shell=True, capture_output=True)
    keys = output.stdout.decode().split()
    return [as_s3_location(location.bucket, k).s3_uri for k in keys]



from datetime import datetime, UTC
from typing import Any, Optional, Self

from yes3.caching import logger
from yes3.caching.base import CacheCore, CachedItemMeta, raise_not_found, _NotSpecified


class MemoryCache(CacheCore):
    def __init__(self, active=True, read_only=False, log_level=None):
        super().__init__(active=active, read_only=read_only, log_level=log_level)
        self._data: dict[str, Any] = {}
        self._meta: dict[str, CachedItemMeta] = {}

    def __contains__(self, key: str):
        if not self.is_active():
            return False
        found = key in self._data
        if found and key not in self._meta:
            raise RuntimeError(f"data exists, but no metadata found, for key '{key}' in {type(self).__name__}")
        return found

    def get(self, key: str, default=_NotSpecified):
        if not self.is_active() or key not in self:
            if default is _NotSpecified:
                raise_not_found(key)
            else:
                return default
        return self._data[key]

    def get_meta(self, key: str) -> CachedItemMeta:
        if not self.is_active() or key not in self:
            raise_not_found(key)
        return self._meta[key]

    def put(
            self,
            key: str,
            obj,
            *,
            update=False,
            meta: Optional[CachedItemMeta] = None,
            log_msg: Optional[str] = None,
    ) -> Self:
        if self.is_read_only():
            raise TypeError('Cache is in read only mode')
        if self.is_active():
            if key in self and not update:
                raise ValueError(f"key '{key}' already exists in cache; use 'update' to overwrite")
            if meta is None:
                meta = CachedItemMeta(key=key, timestamp=datetime.now(UTC), path=None, size=None)
            self._meta[key] = meta
            self._data[key] = obj
        else:
            logger.warning(f"WARNING: {type(self).__name__} is not active")
        if log_msg:
            logger.warn("Log messages are not persisted in MemoryCache")
            logger.info(log_msg)
        return self

    def remove(self, key: str, log_msg: Optional[str] = None) -> Self:
        if self.is_active():
            if key in self:
                if self.is_read_only():
                    raise TypeError('Cache is in read only mode')
                self._data.pop(key)
                self._meta.pop(key)
        else:
            logger.warning(f"WARNING: {type(self).__name__} is not active")
        return self

    def keys(self) -> list[Any]:
        if not self.is_active():
            return []
        else:
            return list(self._data.keys())

    def clear(self, force=False) -> Self:
        if self.is_active():
            if len(self.keys()) > 0:
                if not force:
                    raise RuntimeError(f'Clearing this {type(self).__name__} requires specifying force=True')
                self._data: dict[str, Any] = {}
                self._meta: dict[str, CachedItemMeta] = {}
        else:
            logger.warning(f"WARNING: {type(self).__name__} is not active")
        return self



from collections import defaultdict
from typing import Iterator, Optional, Self

from yes3 import S3Location
from yes3.caching import logger
from yes3.caching.base import CacheCore, CachedItemMeta, check_meta_mismatches, raise_not_found, _NotSpecified


class MultiCache(CacheCore):
    def __init__(
            self,
            caches: list[CacheCore],
            left_to_right_priority=True,
            sync_all=False,
            active=True,
            read_only=False,
            log_level=None,
    ):
        super().__init__(active=active, read_only=read_only, log_level=log_level)
        if left_to_right_priority:
            self._caches = list(caches)
        else:
            self._caches = list(caches[::-1])
        self._sync_all = sync_all

    def __iter__(self) -> Iterator[CacheCore]:
        return iter(self._caches)

    def activate(self) -> Self:
        super().activate()
        for cache in self:
            cache.activate()
        return self

    def deactivate(self) -> Self:
        super().deactivate()
        for cache in self:
            cache.deactivate()
        return self

    def is_active(self) -> bool:
        return super().is_active() and any(cache.is_active() for cache in self)

    def is_read_only(self) -> bool:
        return super().is_read_only() or all(cache.is_read_only() for cache in self)

    def set_read_only(self, value: bool) -> Self:
        super().set_read_only(value)
        for cache in self:
            cache.set_read_only(value)
        return self

    def add_cache(self, cache: CacheCore, index=-1) -> Self:
        if index is not None and index >= 0:
            self._caches.insert(index, cache)
        else:
            self._caches.append(cache)
        return self

    def subcache(self, *args, **kwargs) -> Self:
        subcaches = [cache.subcache(*args, **kwargs) for cache in self]
        return type(self)(subcaches, sync_all=self._sync_all, active=self.is_active(), read_only=self.is_read_only())

    def __contains__(self, key: str):
        for cache in self:
            if key in cache:
                return True
        return False

    def get(self, key: str, default=_NotSpecified, sync=None):
        if sync is None:
            sync = self._sync_all
        result = _NotSpecified
        for cache in self:
            if key in cache:
                result = cache.get(key)
                break
        if result is _NotSpecified:
            if default is _NotSpecified:
                raise_not_found(key)
            else:
                result = default
        elif sync:
            for cache in self:
                if cache.is_read_only():
                    continue
                if key not in cache:
                    cache.put(key, result)
        return result

    def get_meta(self, key) -> CachedItemMeta:
        meta = None
        for cache in self:
            if key in cache:
                c_meta = cache.get_meta(key)
                if meta is None:
                    meta = c_meta
                elif meta != c_meta:
                    logger.warning(f"WARNING: meta data mismatch in caches for '{key}'")
        if meta is None:
            raise_not_found(key)
        return meta

    def check_meta_mismatches(self, key=None) -> dict[str, tuple[CachedItemMeta, ...]]:
        return check_meta_mismatches(self._caches, key=key)

    def compare_all_metadata(self) -> dict[str, dict[str, dict]]:
        metadata = defaultdict(dict)
        for key in self.keys():
            for i, cache in enumerate(self):
                cache_key = f'Cache {i + 1}'
                if hasattr(cache, 'path'):
                    if isinstance(cache.path, S3Location):
                        cache_key += f' ({cache.path.s3_uri})'
                    else:
                        cache_key += f' ({cache.path})'
                metadata[key][cache_key] = cache.get_meta(key).to_dict() if key in cache else None
        return dict(metadata)

    def put(
            self,
            key: str,
            obj,
            *,
            update=False,
            meta: Optional[CachedItemMeta] = None,
            log_msg: Optional[str] = None,
    ) -> Self:
        if self.is_read_only():
            raise TypeError('Cache is in read only mode')
        if not self.is_active():
            logger.warning(f"WARNING: {type(self).__name__} is not active")
            return self
        for cache in self:
            if cache.is_read_only():
                continue
            cache.put(key, obj, update=update, meta=meta, log_msg=log_msg)
            meta = cache.get_meta(key)
            mismatch = self.check_meta_mismatches(key)
            if mismatch and not update:
                logger.warning(f"WARNING: Metadata mismatch for '{key}'. Use update=True to sync across caches.")
            if not self._sync_all and not mismatch:
                break
        return self

    def remove(self, key: str, log_msg: Optional[str] = None) -> Self:
        if self.is_active():
            for cache in self:
                if key in cache:
                    cache.remove(key, log_msg=log_msg)
        else:
            logger.warning(f"WARNING: {type(self).__name__} is not active")
        return self

    def keys(self) -> list[str]:
        if not self.is_active():
            return []
        else:
            keys = []
            for cache in self:
                keys.extend(cache.keys())
            return list(set(keys))

    def sync_now(self) -> Self:
        for key in self.keys():
            obj = None
            meta = None
            for cache in self:
                if key not in cache:
                    if obj is None:
                        obj = self.get(key, sync=False)
                        meta = self.get_meta(key)
                    cache.put(key, obj, meta=meta)
                else:
                    mismatch = self.check_meta_mismatches(key)
                    if mismatch:
                        raise RuntimeError(f"Metadata mismatch for '{key}'")
        return self

    def sync_always(self):
        self._sync_all = True
        self.sync_now()

    def rebuild(self) -> Self:
        for cache in self:
            if hasattr(cache, 'rebuild'):
                cache.rebuild()
        return self

    def __repr__(self):
        return f"{type(self).__name__}({', '.join([str(cache) for cache in self])})"



import json
import logging
from abc import ABCMeta, abstractmethod
from collections.abc import Callable
from datetime import datetime, UTC
from pathlib import Path
from typing import Iterable, Iterator, Optional, Self

from yes3 import s3, S3Location
from yes3.utils.logs import check_level, get_logger

logger = get_logger('caching', level=logging.WARNING)

_NotSpecified = object()

LOG_FILENAME = 'logged_messages.json'  # Default file name for manually logged notes in cache


def raise_not_found(key) -> KeyError:
    raise KeyError(f"key '{key}' not found in cache")


class CachedItemMeta:
    _ts_format = '%Y-%m-%d %H:%M:%S.%f %z'

    def __init__(
            self,
            key: Optional[str] = None,
            path: Optional[str] = None,
            size: Optional[int] = None,
            timestamp: Optional[datetime] = None,
            load_path: Optional[str | Path | S3Location] = None,
    ):
        self.key = key
        self.path = path
        self.size = size
        if isinstance(timestamp, float):
            self.timestamp = datetime.fromtimestamp(timestamp, UTC)
        elif isinstance(timestamp, str):
            self.timestamp = datetime.strptime(timestamp, self._ts_format)
        else:
            self.timestamp = timestamp
        self.load_path = load_path

    def is_loaded(self) -> bool:
        return self.key is not None

    def load(self) -> Self:
        if isinstance(self.load_path, S3Location) or (isinstance(self.load_path, str) and s3.is_s3_url(self.load_path)):
            meta = s3.read(self.load_path, file_type='json')
        else:
            with open(self.load_path, 'r') as f:
                meta = json.load(f)
        return type(self)(**meta)

    def to_dict(self) -> dict:
        if not self.is_loaded():
            self.load()
        return {
            'key': self.key,
            'path': self.path,
            'size': self.size,
            'timestamp': self.timestamp.strftime(self._ts_format) if self.timestamp else None,
        }

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join(f'{k}={v!r}' for k, v in self.to_dict().items())})"


class CacheCore(metaclass=ABCMeta):
    def __init__(self, active=True, read_only=False, log_level=None, log_filename=LOG_FILENAME):
        self._read_only = read_only
        self._active = active
        self._log_level = None
        self.set_log_level(log_level)
        self._log_filename = log_filename

    def set_log_level(self, level) -> Self:
        if level != self._log_level:
            logger.info(f"Setting log level to {level}")
        self._log_level = level
        if level is not None:
            logger.setLevel(check_level(level))
        return self

    def get_log_level(self) -> Optional[int]:
        return self._log_level

    @abstractmethod
    def __contains__(self, key):
        pass

    @abstractmethod
    def get(self, key, default=_NotSpecified):
        pass

    @abstractmethod
    def get_meta(self, key) -> CachedItemMeta:
        pass

    @abstractmethod
    def put(
            self,
            key,
            obj,
            *,
            update=False,
            meta: Optional[CachedItemMeta] = None,
            log_msg: Optional[str] = None,
    ) -> Self:
        pass

    @abstractmethod
    def remove(self, key, log_msg: Optional[str] = None):
        pass

    @abstractmethod
    def keys(self):
        pass

    def __getitem__(self, key: str):
        return self.get(key)

    def __setitem__(self, key: str, obj) -> None:
        self.put(key, obj)

    def __delitem__(self, key: str) -> None:
        self.remove(key)

    def is_active(self) -> bool:
        return self._active

    def activate(self):
        self._active = True
        return self

    def deactivate(self):
        self._active = False
        return self

    def is_read_only(self) -> bool:
        return self._read_only

    def set_read_only(self, value: bool) -> Self:
        self._read_only = value
        return self

    def update(self, key: str, obj, meta: Optional[CachedItemMeta] = None, log_msg: Optional[str] = None):
        if key not in self:
            raise_not_found(key)
        self.put(key, obj, update=True, meta=meta, log_msg=log_msg)

    def pop(self, key: str, default=_NotSpecified, log_msg: Optional[str] = None):
        obj = self.get(key, default=default)
        self.remove(key, log_msg=log_msg)
        return obj

    def list(self) -> dict[str, CachedItemMeta]:
        items_meta = {}
        for key in self.keys():
            items_meta[key] = self.get_meta(key)
        return items_meta

    def subcache(self, *args, **kwargs) -> Self:
        raise NotImplementedError(f"`subcache` method is not defined for class {type(self).__name__}")

    def write_log_msg(self, msg: str):
        raise NotImplementedError(f"`write_log_msg` method is not implemented for class {type(self).__name__}")

    def read_log(self):
        raise NotImplementedError(f"`read_log` method is not implemented for class {type(self).__name__}")


class CacheReaderWriter(metaclass=ABCMeta):
    @abstractmethod
    def read(self, key: str):
        pass

    @abstractmethod
    def get_meta(self, key: str) -> CachedItemMeta:
        pass

    @abstractmethod
    def write(self, key: str, obj, meta=None) -> CachedItemMeta:
        pass

    @abstractmethod
    def delete(self, key: str, meta_only=False):
        pass


class CacheCatalog(metaclass=ABCMeta):
    @abstractmethod
    def contains(self, key: str):
        pass

    @abstractmethod
    def add(self, key: str, info: CachedItemMeta):
        pass

    @abstractmethod
    def get(self, key: str) -> CachedItemMeta:
        pass

    @abstractmethod
    def remove(self, key: str):
        pass

    @abstractmethod
    def keys(self):
        pass

    @abstractmethod
    def items(self):
        pass

    @abstractmethod
    def rebuild(self):
        pass


_CatalogT = dict[str, CachedItemMeta]
_CatalogBuilderT = Callable[[], _CatalogT]


class CacheDictCatalog(CacheCatalog):
    def __init__(
            self,
            catalog: Optional[dict[str, CachedItemMeta]] = None,
            catalog_builder: Optional[_CatalogBuilderT] = None,
    ):
        self._catalog = catalog
        if catalog_builder is None:
            catalog_builder = dict
        self._build_catalog = catalog_builder
        if self._catalog is None:
            self.rebuild()

    def rebuild(self):
        self._catalog = self._build_catalog().copy()

    def contains(self, key: str):
        return str(key) in self._catalog

    def add(self, key: str, meta: CachedItemMeta):
        self._catalog[str(key)] = meta

    def get(self, key: str) -> CachedItemMeta:
        meta = self._catalog[str(key)]
        if not meta.is_loaded():
            meta = meta.load()
        return meta

    def remove(self, key: str):
        self._catalog.pop(str(key))

    def keys(self):
        return list(self._catalog.keys())

    def items(self) -> Iterator[tuple[str, CachedItemMeta]]:
        return iter(self._catalog.items())


class Cache(CacheCore, metaclass=ABCMeta):
    def __init__(
            self,
            catalog: CacheCatalog,
            reader_writer: CacheReaderWriter,
            active=True,
            read_only=False,
            log_level=None,
    ):
        super().__init__(active=active, read_only=read_only, log_level=log_level)
        self._catalog = catalog
        self._reader_writer = reader_writer

    @classmethod
    @abstractmethod
    def create(cls, *args, **kwargs):
        pass

    def __contains__(self, key: str) -> bool:
        if not self.is_active():
            return False
        return self._catalog.contains(key)

    def get(self, key: str, default=_NotSpecified):
        if not self.is_active() or key not in self:
            if default is _NotSpecified:
                raise_not_found(key)
            else:
                return default
        return self._reader_writer.read(key)

    def get_meta(self, key: str) -> CachedItemMeta:
        if not self.is_active() or key not in self:
            raise_not_found(key)
        return self._catalog.get(key)

    def put(
            self,
            key: str,
            obj,
            *,
            update=False,
            meta: Optional[CachedItemMeta] = None,
            log_msg: Optional[str] = None,
    ) -> Self:
        if self.is_read_only():
            raise TypeError('Cache is in read only mode')
        if self.is_active():
            if key in self and not update:
                raise ValueError(f"key '{key}' already exists in cache; use 'update' to overwrite")
            meta = self._reader_writer.write(key, obj, meta=meta)
            self._catalog.add(key, meta)
            if log_msg:
                self.write_log_msg(log_msg)
        else:
            logger.info(f'WARNING: {type(self).__name__} is not active')
        return self

    def remove(self, key: str, meta_only=False, log_msg: Optional[str] = None) -> Self:
        if self.is_active() and key in self:
            if self.is_read_only():
                raise TypeError('Cache is in read only mode')
            self._catalog.remove(key)
            self._reader_writer.delete(key, meta_only=meta_only)
            if log_msg:
                self.write_log_msg(log_msg)
        return self

    def remove_meta(self, key: str, log_msg: Optional[str] = None) -> Self:
        return self.remove(key, meta_only=True, log_msg=log_msg)

    def keys(self) -> list[str]:
        if not self.is_active():
            return []
        else:
            return list(self._catalog.keys())

    def rebuild(self) -> Self:
        self._catalog.rebuild()
        return self

    def _repr_params(self) -> list[str]:
        params = [f'{len(self.keys())} items']
        if not self.is_active():
            params.append('NOT ACTIVE')
        if self.is_read_only():
            params.append('READ ONLY')
        return params

    def __repr__(self):
        return f"{type(self).__name__}({', '.join(self._repr_params())})"


def check_meta_mismatches(caches: Iterable[CacheCore], key=None) -> dict[str, tuple[CachedItemMeta, ...]]:
    if key is not None and not isinstance(key, str):
        raise TypeError('key is not a string')
    for cache in caches:
        if not isinstance(cache, CacheCore):
            raise TypeError('caches must be an iterable containing Cache instances')
    mismatches = {}
    if key is None:
        keys = set(k for cache in caches for k in cache.keys())
    else:
        keys = [key]
    for k in keys:
        metas = [cache.get_meta(k) for cache in caches if k in cache]
        if len(metas) > 1:
            first_meta = metas[0]
            if any(meta != first_meta for meta in metas[1:]):
                mismatches[k] = tuple(metas)
    return mismatches


class Serializer(metaclass=ABCMeta):
    default_ext = None

    def __init__(self, ext=None):
        if ext is None:
            self.ext = self.default_ext
        else:
            self.ext = ext

    @abstractmethod
    def read(self, path):
        pass

    @abstractmethod
    def write(self, path, obj):
        pass



import logging
import sys
from pathlib import Path
from typing import Optional


def check_level(level: int | str) -> int:
    """Check if the logging level is valid and return the corresponding integer value."""
    if isinstance(level, str):
        level_map = logging.getLevelNamesMapping()
        try:
            level = level_map[level.upper()]
        except KeyError:
            raise ValueError(f"Invalid logging level '{level}', must be one of {list(level_map.keys())}")
    if not isinstance(level, int):
        raise TypeError(f"Logging level must be an int or str, got {type(level)}")
    return level


def get_log_format(asctime=True, name=True, levelname=True) -> str:
    """Get the default log format string."""
    parts = []
    if asctime:
        parts.append("%(asctime)s")
    if name:
        parts.append("%(name)s")
    if levelname:
        parts.append("%(levelname)s")
    prefix = " - ".join(parts)
    if prefix:
        fmt = f"[{prefix}] "
    else:
        fmt = ""
    return fmt + "%(message)s"


def get_logger(
        name: Optional[str] = None,
        level: int | str = logging.WARNING,
        fmt: Optional[str] = None,
        path: Optional[str | Path] = None,
        stdout: Optional[bool] = None,
) -> logging.Logger:
    """Get a logger with the specified name and logging level."""
    level = check_level(level)
    logger = logging.getLogger(name)
    if not fmt:
        fmt = get_log_format()
    if path:
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        file_handler = logging.FileHandler(path)
        file_handler.setLevel(level)
        formatter = logging.Formatter(fmt)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
        if stdout is True:
            stdout_handler = logging.StreamHandler(sys.stdout)
            stdout_handler.setLevel(level)
            stdout_handler.setFormatter(formatter)
            logger.addHandler(stdout_handler)
    else:
        logging.basicConfig(level=level, format=fmt, stream=(sys.stdout if stdout else None))
    return logger



from datetime import datetime
from functools import wraps
from inspect import signature


def timeit_opt(func=None, *, default=False):
    assert func is None or callable(func), 'timeit_opt parameters must be passed as keyword arguments'

    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            if 'timeit' in signature(f).parameters:
                timeit = kwargs.get('timeit', default)
            else:
                timeit = kwargs.pop('timeit', default)
            start = datetime.now()
            result = f(*args, **kwargs)
            if timeit:
                print(f'{f.__name__} elapsed time: {datetime.now() - start}')
            return result
        return wrapper

    if func is None:
        return decorator
    else:
        return decorator(func)
